"""Utility module."""
import datetime
import typing as t

import numpy as np
import pint
import xarray as xr

from ..constants import K, MM
from ..units import ureg


def utcnow() -> str:
    """Get current UTC time.
    
    Returns:
        ISO 8601 formatted UTC timestamp.
    """
    return datetime.datetime.utcnow().replace(microsecond=0).isoformat()


def number_density(p: pint.Quantity, t: pint.Quantity) -> pint.Quantity:
    """Compute air number density from air pressure and air temperature.
    
    Args:
        p: Air pressure.
        t: Air temperature.
    
    Returns:
        Number density.
    
    Notes:
        The air number density is computed according to the ideal gas law:

        $$
        n = \\frac{p}{k_B T}
        $$

        where $p$ is the air pressure, $k_B$ is the Boltzmann constant, and 
        $T$ is the air temperature.
    """
    return (p / (K * t)).to_base_units()


def molar_mass(molecules: t.List[str]) -> xr.DataArray:
    return xr.DataArray(
        data=np.array([MM[m] for m in molecules]),
        coords={"m": ("m", molecules)},
        attrs={
            "long_name": "molar mass",
            "standard_name": "molar_mass",
            "units": "g/mol"
        },
    )


def to_m_suffixed_data(da: xr.DataArray) -> xr.Dataset:
    """
    Takes a data array with a `m` coordinate and returns a dataset with
    a data variable for each molecule.

    Args:
        da: Data array with a `m` coordinate. Must be named.
    
    Returns:
        Dataset with a data variable for each molecule.
    """
    data_var_name = da.name
    data_var_attrs = da.attrs
    molecules = da.m.values
    ds = xr.Dataset()
    for m in molecules:
        ds[f"{data_var_name}_{m}"] = da.sel(m=m, drop=True)
        ds[f"{data_var_name}_{m}"].attrs = data_var_attrs
    return ds


def air_molar_mass_from_mass_fraction(
    y: xr.DataArray
) -> xr.DataArray:
    r"""
    Compute the air molar mass from the of air constituents mass fractions.
    
    Args:
        y: Mass fraction as a function of molecule (`m`) and altitude (`z`).
    
    Returns:
        Air molar mass as a function of altitude (`z`).
    
    Notes:
        The air molar mass is computed according to the following equation:

        $$
        m_{\mathrm{air}} (z) = \left(
            \sum_{\mathrm{M}} \frac{
                y_{\mathrm{M}} (z)
            }{
                m_{\mathrm{M}}
            }
        \right)^{-1}
        $$

        where:

        * $y_{\mathrm{M}} (z)$ is the mass fraction of molecule M at altitude
          $z$,
        * $m_{\mathrm{M}}$ is the molar mass of molecule M.
    """
    # compute molar masses along molecular species
    molecules = y.m.values
    mm = xr.DataArray(
        data=[MM[m] for m in molecules],
        coords={"m": ("m", molecules)},
    )

    # compute air molar mass
    mm_inv = 1 / mm
    weighted_mean_mm_inv = (mm_inv * y).sum(dim="m") / y.sum(dim="m")
    mm_air = 1 / weighted_mean_mm_inv
    mm_air.attrs.update({"units": "g/mol"})
    return mm_air



def mass_fraction_to_mole_fraction3(
    y: xr.DataArray,
    m_air: pint.Quantity = 28.9644 * ureg.g / ureg.mole,
) -> xr.DataArray:
    r"""
    Convert mass fractions to mole fractions.
    
    Args:
        y: Mass fraction as a function of molecule (`m`) and altitude (`z`).
        m_air: Molar mass of air. Defaults to 28.9644 g/mol.

    Returns:
        Volume fraction as a function of molecule (`m`) and altitude (`z`).
    
    Notes:
        The mole fraction of molecule M at altitude $z$ is computed according 
        to the following equation:

        $$
        x_{\mathrm{M}} (z) = \frac{
            y_{\mathrm{M}} (z)
        }{
            m_{\mathrm{M}}
        }
        m_{\mathrm{air}} (z)
        $$

        where:

        * $y_{\mathrm{M}} (z)$ is the mass fraction of molecule $M$ at 
          altitude $z$,
        * $m_{\mathrm{M}}$ is the molar mass of molecule $M$,
        * $m_{\mathrm{air}}$ is the air molar mass (`m_air`).
    """
    # compute molar masses of each molecular species
    mm = molar_mass(molecules=y.m.values.tolist())

    # compute mole fraction
    x = (m_air.m_as("g/mol") * y / mm).rename("x")
    x.attrs.update(
        {
            "long_name": "Mole fraction",
            "standard_name": "mole_fraction",
            "units": "dimensionless"
        }
    )

    return x



"""Test cases for units module."""
import numpy as np
import pint
import pytest
import xarray as xr
from numpy.typing import ArrayLike

from joseki.units import ureg, to_quantity


@pytest.fixture
def dataset() -> xr.Dataset:
    """Fixture for working with a dummy dataset."""
    return xr.Dataset(
        data_vars={
            "x": ("t", np.random.random(50), {"units": "m"}),
            "y": ("t", np.random.random(50)),
        },
        coords={"t": ("t", np.linspace(0, 10, 50), {"units": "s"})},
    )

def test_to_quantity_invalid_type():
    """Raises a TypeError."""
    with pytest.raises(NotImplementedError):
        to_quantity("not a quantity")

def test_to_quantity_quantity():
    """Returns the same quantity."""
    q = ureg.Quantity(1.0, "m")
    assert to_quantity(q) is q

    assert to_quantity(q, units="cm").units == ureg.Unit("cm")

def test_to_quantity_dict():
    """Returns a quantity."""
    d = {"value": 1.0, "units": "m"}
    assert isinstance(to_quantity(d), pint.Quantity)

@pytest.mark.parametrize(
    "value",
    [
        1,
        1.0,
        [1.0, 2.0],
        [[1.0, 2.0], [3.0, 4.0]],
        np.array([1.0, 2.0])
    ],
)
def test_to_quantity_array_like(value: ArrayLike):
    """Returns a dimensionless quantity."""
    assert to_quantity(value).units == ureg.dimensionless

def test_to_quantity_da(dataset: xr.Dataset):
    """Returns a quantity."""
    assert isinstance(to_quantity(dataset.x), pint.Quantity)

def test_to_quantity_da_units(dataset: xr.Dataset):
    """Returns a quantity with cm units."""
    q = to_quantity(dataset.x, units="cm")
    assert q.units == ureg.Unit("cm")

def test_to_quantity_da_no_units_raise(dataset: xr.Dataset):
    """Returns a quantity."""
    with pytest.raises(ValueError):
        to_quantity(dataset.y)


def test_to_quantity_da_no_units_no_raise(dataset: xr.Dataset):
    """Returns a quantity."""
    assert isinstance(to_quantity(dataset.y, units="m"), pint.Quantity)



"""MIPAS atmosphere thermophysical profiles.

[Remedios et al. (2007)](bibliography.md#Remedios+2007) define a set of 5 
"standard atmospheres" representing the atmosphere at different latitudes and 
seasons or times of day:

* midlatitude day
* midlatitude night
* polar winter
* polar summer
* tropical


MIPAS standard atmospheres were intended to provide an updated set of pro-
files for characteristic atmospheric states such as the
[AFGL Atmospheric constituent profiles](bibliography.md#Anderson+1986).
"""
import enum
import importlib_resources
import logging
import typing as t

import numpy as np
import pint
import xarray as xr
from attrs import define
from scipy.constants import physical_constants

from ..units import ureg
from .core import Profile, interp, DEFAULT_METHOD
from .factory import factory
from .schema import history, schema

logger = logging.getLogger(__name__)

SOURCE = "Combination of model and observational data"

REFERENCE = (
    "Remedios, John J. et al. “MIPAS reference atmospheres and comparisons to "
    "V4.61/V4.62 MIPAS level 2 geophysical data sets.” Atmospheric Chemistry "
    "and Physics 7 (2007): 9973-10017. DOI: 10.5194/ACPD-7-9973-2007"
)

INSTITUTION = "EOS, Space Research Centre, Leicester, U.K."

URL = "https://eodg.atm.ox.ac.uk/RFM/atm/"

URL_DATE = "2022-12-12"

CFC_FORMULAE = {
    "CCl3F": ("Freon-11", "F11", "R-11", "CFC-11"),
    "CCl2F2": ("Freon-12", "F12", "R-12", "CFC-12"),
    "CClF3": ("Freon-13", "F13", "R-13", "CFC-13"),
    "CF4": ("Freon-14", "F14", "CFC-14"),
    "CHClF2": ("Freon-22", "F22", "HCFC-22"),
    "CHCl2F": ("Freon-21", "F21", "HCFC-21"),
    "C2Cl3F3": ("Freon-113", "F113", "CFC-113"),
    "C2Cl2F4": ("Freon-114", "F114", "CFC-114"),
}

# Boltzmann constant
K = ureg.Quantity(*physical_constants["Boltzmann constant"][:2])


def to_chemical_formula(name: str) -> str:
    """Convert to chemical formula.
    
    Args:
        name: Molecule name.

    Returns:
        Molecule formula.
    
    Notes:
        If molecule name is unknown, returns name unchanged.
    """
    try:
        return translate_cfc(name)
    except ValueError:
        return name


def translate_cfc(name: str) -> str:
    """Convert chlorofulorocarbon name to corresponding chemical formula.

    Args:
        name: Chlorofulorocarbon name.

    Returns:
        Chlorofulorocarbon chemical formula.

    Raises:
        ValueError: If the name does not match a known chlorofulorocarbon.
    """
    for formula, names in CFC_FORMULAE.items():
        if name in names:
            return formula
    raise ValueError("Unknown chlorofulorocarbon {name}")


class Identifier(enum.Enum):
    """MIPAS atmosphere thermophysical profile identifier enumeration."""

    MIDLATITUDE_DAY = "midlatitude_day"
    MIDLATITUDE_NIGHT = "midlatitude_night"
    POLAR_WINTER = "polar_winter"
    POLAR_SUMMER = "polar_summer"
    TROPICAL = "tropical"


def parse_units(s: str) -> str:
    """Parse units."""
    if s.startswith("[") and s.endswith("]"):
        units = s[1:-1]
        if units == "mb":
            return "millibar"
        else:
            return units
    else:
        raise ValueError(f"Cannot parse units '{s}'")


def parse_var_name(n: str) -> str:
    """Parse variable name."""
    translate = {"HGT": "z", "PRE": "p", "TEM": "t"}
    if n in translate.keys():
        return translate[n]
    else:
        return to_chemical_formula(n)


def parse_var_line(s: str) -> t.Tuple[str, str]:
    """Parse a line with the declaration of a variable and its units."""
    parts = s[1:].strip().split()
    if len(parts) == 2:
        var_name, units_s = parts
    elif len(parts) == 3:
        var_name, _, units_s = parts
    else:
        raise ValueError(f"Invalid line format: {s}")
    var = parse_var_name(var_name)
    units = parse_units(units_s)
    return var, units


def parse_values_line(s: str) -> t.List[str]:
    """Parse a line with numeric values."""
    if "," in s:  # delimiter is comma and whitespace combined
        s_strip = s.strip()
        if s_strip[-1] == ",":
            s_strip = s_strip[:-1]
        return [x.strip() for x in s_strip.split(",")]
    else:  # delimiter is whitespace
        return s.split()


def parse_content(lines: t.List[str]) -> t.Dict[str, pint.Quantity]:
    """Parse lines."""
    logger.debug("Parsing file content")
    iterator = iter(lines)
    line = next(iterator)

    quantities: t.Dict[str, pint.Quantity] = {}

    def _add_to_quantities(quantity: pint.Quantity, name: str) -> None:
        if name not in ["z", "p", "t", "n"]:
            name = f"x_{name}"
        
        if quantity.check(""):
            quantities[name] = quantity.to("dimensionless")
        else:
            quantities[name] = quantity

    var: str = ""
    units: str = ""
    values: t.List[str] = []
    while line != "*END":
        if line.startswith("!"):
            pass  # this is a comment, ignore the line
        elif line.startswith("*"):
            # convert previously read values (if any) and units to quantity
            if len(values) > 0:
                quantity = ureg.Quantity(
                    np.array(values, dtype=float),
                    units,
                )
                _add_to_quantities(quantity=quantity, name=var)

            # this is a variable line, parse variable name and units
            var, units = parse_var_line(line)

            # following lines are the variables values so prepare a variable
            # to store the values
            values = []
        else:
            if "!" in line:
                # this the line with the number of profile levels, ignore it
                pass
            else:
                # this line contains variable values
                values += parse_values_line(line)
        line = next(iterator)

    # include last array of values before the '*END' line
    quantity = ureg.Quantity(np.array(values, dtype=float), units)
    _add_to_quantities(quantity=quantity, name=var)

    return quantities


def read_file_content(identifier: Identifier) -> str:
    """
    Read data file content.

    Args:
        identifier: Atmospheric profile identifier.
            See 
            [`Identifier`](reference.md#src.joseki.profiles.mipas_2007.Identifier) 
            for possible values.

    Returns:
        file content, URL, URL date.
    """
    package = "joseki.data.mipas_2007"
    file = f"{identifier.value}.atm"
    logger.debug(f"Reading file {file}")
    return importlib_resources.files(package).joinpath(file).read_text()


def get_dataset(identifier: Identifier) -> xr.Dataset:
    """Read MIPAS reference atmosphere data files into an xarray.Dataset.

    Args:
        identifier: Atmospheric profile identifier.
            See 
            [`Identifier`](reference.md#src.joseki.profiles.mipas_2007.Identifier) 
            for possible values.

    Returns:
        Atmospheric profile.
    """
    content = read_file_content(identifier=identifier)
    quantities = parse_content(content.splitlines())

    # Coordinates
    coords = {"z": quantities.pop("z")}

    # Data variables
    data_vars = {}
    p = quantities.pop("p")
    data_vars["p"] = p
    t = quantities.pop("t")
    data_vars["t"] = t
    n = p / (K * t)  # perfect gas equation
    data_vars["n"] = n
    data_vars.update(quantities)

    logger.debug("data variables: %s", data_vars.keys())

    # Attributes
    pretty_id = identifier.value.replace("_", " ")
    pretty_title = f"MIPAS {pretty_id} Reference Atmosphere"
    attrs = {
        "Conventions": "CF-1.10",
        "history": history(),
        "title": pretty_title,
        "source": SOURCE,
        "institution": INSTITUTION,
        "references": REFERENCE,
        "url": URL,
        "urldate": URL_DATE,
    }

    # Dataset
    ds = schema.convert(
        data_vars=data_vars,
        coords=coords,
        attrs=attrs,
    )

    return ds


def to_dataset(
    identifier: Identifier,
    z: t.Optional[pint.Quantity] = None,
    method: t.Optional[t.Mapping[str, str]] = None,
    conserve_column: bool = False,
    **kwargs: t.Any,
) -> xr.Dataset:
    """Helper for `Profile.to_dataset` method"""
    # no kwargs are expected
    if len(kwargs) > 0:  # pragma: no cover
        logger.warning("Unexpected keyword arguments: %s", kwargs)

    # get original MIPAS midlatitude day reference atmosphere
    logger.debug("Get original MIPAS midlatitude day reference atmosphere")
    ds = get_dataset(identifier=identifier)

    # Interpolate to new vertical grid if necessary
    if z is not None:
        method = DEFAULT_METHOD if method is None else method
        ds = interp(
            ds=ds,
            z_new=z,
            method=method,
            conserve_column=conserve_column,
        )
        return ds
    else:
        return ds


@factory.register("mipas_2007-midlatitude_day")
@define
class MIPASMidlatitudeDay(Profile):
    """MIPAS midlatitude day reference atmosphere."""

    def to_dataset(
        self,
        z: t.Optional[pint.Quantity] = None,
        interp_method: t.Optional[t.Mapping[str, str]] = None,
        conserve_column: bool = False,
        **kwargs: t.Any,
    ) -> xr.Dataset:
        return to_dataset(
            identifier=Identifier.MIDLATITUDE_DAY,
            z=z,
            method=interp_method,
            conserve_column=conserve_column,
            **kwargs,
        )


@factory.register("mipas_2007-midlatitude_night")
@define
class MIPASMidlatitudeNight(Profile):
    """MIPAS Midlatitude night reference atmosphere."""

    def to_dataset(
        self,
        z: t.Optional[pint.Quantity] = None,
        interp_method: t.Optional[t.Mapping[str, str]] = None,
        conserve_column: bool = False,
        **kwargs: t.Any,
    ) -> xr.Dataset:
        return to_dataset(
            identifier=Identifier.MIDLATITUDE_NIGHT,
            z=z,
            method=interp_method,
            conserve_column=conserve_column,
            **kwargs,
        )


@factory.register("mipas_2007-polar_summer")
@define
class MIPASPolarSummer(Profile):
    """MIPAS Polar summer reference atmosphere."""

    def to_dataset(
        self,
        z: t.Optional[pint.Quantity] = None,
        interp_method: t.Optional[t.Mapping[str, str]] = None,
        conserve_column: bool = False,
        **kwargs: t.Any,
    ) -> xr.Dataset:
        return to_dataset(
            identifier=Identifier.POLAR_SUMMER,
            z=z,
            method=interp_method,
            conserve_column=conserve_column,
            **kwargs,
        )


@factory.register("mipas_2007-polar_winter")
@define
class MIPASPolarWinter(Profile):
    """MIPAS Polar winter reference atmosphere."""

    def to_dataset(
        self,
        z: t.Optional[pint.Quantity] = None,
        interp_method: t.Optional[t.Mapping[str, str]] = None,
        conserve_column: bool = False,
        **kwargs: t.Any,
    ) -> xr.Dataset:
        return to_dataset(
            identifier=Identifier.POLAR_WINTER,
            z=z,
            method=interp_method,
            conserve_column=conserve_column,
            **kwargs,
        )


@factory.register("mipas_2007-tropical")
@define
class MIPASTropical(Profile):
    """MIPAS Tropical reference atmosphere."""

    def to_dataset(
        self,
        z: t.Optional[pint.Quantity] = None,
        interp_method: t.Optional[t.Mapping[str, str]] = None,
        conserve_column: bool = False,
        **kwargs: t.Any,
    ) -> xr.Dataset:
        return to_dataset(
            identifier=Identifier.TROPICAL,
            z=z,
            method=interp_method,
            conserve_column=conserve_column,
            **kwargs,
        )



"""Test cases for the accessor module."""
import numpy as np
import pint
import pytest
from numpy.testing import assert_approx_equal

import joseki
from joseki.accessor import _scaling_factor
from joseki.units import to_quantity
from joseki.units import ureg


def test_column_number_density_integration():
    """Same result is returned when profile is represented in cells or not."""
    ds1 = joseki.make("mipas_2007-midlatitude_day")
    ds2 = joseki.make("mipas_2007-midlatitude_day", represent_in_cells=True)
    assert_approx_equal(
        ds1.joseki.column_number_density["CO2"].m,
        ds2.joseki.column_number_density["CO2"].m,
        significant=5,
    )

@pytest.mark.parametrize(
    "identifier",
    ["ussa_1976", "afgl_1986-us_standard", "mipas_2007-midlatitude_day"]
)
def test_column_mass_density(identifier: str):
    """Returns a dictionary."""
    ds = joseki.make(identifier)
    assert isinstance(ds.joseki.column_mass_density, dict)


@pytest.mark.parametrize(
    "identifier, represent_in_cells, expected",
    [
        ("afgl_1986-us_standard", True, 14.27 * ureg.kg / ureg.m**2),
        ("afgl_1986-us_standard", False, 14.38 * ureg.kg / ureg.m**2),
        ("mipas_2007-midlatitude_day", True, 19.33 * ureg.kg / ureg.m**2),
        ("mipas_2007-midlatitude_day", False, 19.51 * ureg.kg / ureg.m**2),
    ]
)
def test_water_vapour_column_mass_density_in_cells(
    identifier: str,
    represent_in_cells: bool,
    expected: pint.Quantity
):
    """Column mass density of water vapor is close to expected values."""
    ds = joseki.make(identifier, represent_in_cells=represent_in_cells)
    water_vapor_amount = ds.joseki.column_mass_density["H2O"].to("kg/m^2")
    assert_approx_equal(water_vapor_amount.m, expected.m, significant=3)


@pytest.mark.parametrize(
    "identifier, represent_in_cells, expected",
    [
        ("afgl_1986-us_standard", True, 348.3 * ureg.dobson_unit),
        ("afgl_1986-us_standard", False, 345.7 * ureg.dobson_unit),
        ("mipas_2007-midlatitude_day", True, 303.4 * ureg.dobson_unit),
        ("mipas_2007-midlatitude_day", False, 301.7 * ureg.dobson_unit)
    ]
)
def test_column_number_density(
    identifier: str,
    represent_in_cells: bool,
    expected: pint.Quantity
):
    """Column number density of ozone matches expected value."""
    ds = joseki.make(identifier, represent_in_cells=represent_in_cells)
    ozone_amount = ds.joseki.column_number_density["O3"].to("dobson_unit")
    assert_approx_equal(ozone_amount.m, expected.m, significant=3)


@pytest.mark.parametrize(
    "identifier", ["afgl_1986-midlatitude_summer", "mipas_2007-midlatitude_day"]
)
def test_number_density_at_sea_level(identifier: str):
    """Number density at sea level of CO2 in matches dataset values."""
    ds = joseki.make(identifier)
    carbon_dioxide_amount = ds.joseki.number_density_at_sea_level["CO2"]
    n = to_quantity(ds.n.isel(z=0))
    x_co2 = to_quantity(ds.x_CO2.isel(z=0))
    expected = (x_co2 * n).to("m^-3")
    value = carbon_dioxide_amount.to("m^-3")
    assert_approx_equal(value.m, expected.m)


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("afgl_1986-us_standard", 0.00590723 * ureg.kg / ureg.m**3),
        ("mipas_2007-midlatitude_day", 0.00901076 * ureg.kg / ureg.m**3)
    ]
)
def test_mass_density_at_sea_level(identifier: str, expected: pint.Quantity):
    """Mass density at sea level of H2O matches expected value."""
    ds = joseki.make(identifier)
    water_vapor_amount = ds.joseki.mass_density_at_sea_level["H2O"].to("kg/m^3")
    assert_approx_equal(water_vapor_amount.m, expected.m, significant=6)


@pytest.mark.parametrize(
    "identifier, expected",
    [
        ("afgl_1986-us_standard", 0.000330 * ureg.dimensionless),
        ("mipas_2007-midlatitude_day", 0.0003685 * ureg.dimensionless)
    ]
)
def test_mole_fraction_at_sea_level(identifier: str, expected):
    """CO2 mole mixing fraction at sea level matches expected value."""
    ds = joseki.make(identifier)
    value = ds.joseki.mole_fraction_at_sea_level["CO2"].to(ureg.dimensionless)
    assert_approx_equal(value.m, expected.m)


def test_scaling_factor():
    """Returns 2.0 when target amount is twice larger than initial amount."""
    initial = 5 * ureg.m
    target = 10 * ureg.m
    assert _scaling_factor(initial_amount=initial, target_amount=target) == 2.0


def test_scaling_factor_zero():
    """Returns 0.0 when target amount and initial amount are both zero."""
    initial = 0.0 * ureg.m
    target = 0.0 * ureg.m
    assert _scaling_factor(initial_amount=initial, target_amount=target) == 0.0


def test_scaling_factor_raises():
    """Raises when the initial amount is zero but not the target amount."""
    initial = 0 * ureg.m
    target = 10 * ureg.m
    with pytest.raises(ValueError):
        _scaling_factor(initial_amount=initial, target_amount=target)


@pytest.mark.parametrize(
    "identifier", ["afgl_1986-us_standard", "mipas_2007-midlatitude_day"]
)
def test_scaling_factors(identifier: str):
    """Scaling factors keys match target amounts keys."""
    ds = joseki.make(identifier)
    target = {
        "H2O": 20.0 * ureg.kg * ureg.m**-2,
        "O3": 350.0 * ureg.dobson_unit,
    }
    factors = ds.joseki.scaling_factors(target=target)
    assert all([k1 == k2 for k1, k2 in zip(target.keys(), factors.keys())])


@pytest.mark.parametrize(
    "identifier", ["afgl_1986-us_standard"]
)
def test_rescale(identifier: str):
    """Scaling factors are applied."""
    ds = joseki.make(identifier)
    factors = {
        "H2O": 0.8,
        "CO2": 1.9,
    }
    initial = ds.copy(deep=True)
    rescaled = ds.joseki.rescale(factors)
    assert all(
        [
            np.allclose(
                rescaled[f"x_{m}"],
                factors[m] * initial[f"x_{m}"],
            )
            for m in factors
        ]
    )


@pytest.mark.parametrize(
    "identifier", ["afgl_1986-us_standard", "mipas_2007-midlatitude_day"]
)
def test_rescale_invalid(identifier: str):
    """A UserWarning is raised if invalid scaling factors are passed."""
    ds = joseki.make(identifier)
    with pytest.raises(ValueError):
        ds.joseki.rescale(
            factors={"O2": 2.0},
            check_x_sum=True,
        )

def test_rescale_to_column_mass_density():
    """Column mass density of H2O matches target value."""
    ds = joseki.make("afgl_1986-midlatitude_summer")
    target = {
        "H2O": 20.0 * ureg.kg * ureg.m**-2,
    }
    rescaled = ds.joseki.rescale_to(target)
    assert_approx_equal(
        rescaled.joseki.column_mass_density["H2O"].m,
        target["H2O"].m,
        significant=3,
    )

@pytest.mark.parametrize(
    "molecules",
    [
        ["H2O"],
        ["H2O", "CO2"],
    ]
)
def test_drop_molecules(molecules):
    """x_M data variable is(are) dropped."""
    ds = joseki.make("afgl_1986-midlatitude_summer")
    assert all([m in ds.joseki.molecules for m in molecules])
    dropped = ds.joseki.drop_molecules(molecules)
    assert all([m not in dropped.joseki.molecules for m in molecules])



"""Test cases for the ussa_1976 module."""
from joseki.profiles.ussa_1976 import USSA1976

def test_ussa_1976():
    """Returns a valid dataset."""
    profile = USSA1976()
    ds = profile.to_dataset()
    assert ds.joseki.is_valid



import tempfile
from pathlib import Path

import numpy as np
import pint
import pytest
from numpy.testing import assert_almost_equal, assert_allclose

import joseki
from joseki import unit_registry as ureg
from joseki.units import to_quantity
from joseki.profiles.cams import (
    from_cams_reanalysis,
    get_molecule_amounts,
    from_archive,
    to_paths,
)


def test_from_cams_reanalysis():
    """CAMS reanalysis data can be transformed into a Joseki profile."""
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
    )
    assert ds.joseki.is_valid
    expected_molecules = ["H2O", "O3"]
    assert all(m in ds.joseki.molecules for m in expected_molecules)


def test_from_cams_reanalysis_molecules_drop():
    """Unrequested molecules are dropped."""
    molecules = ["H2O"]
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        molecules=molecules,
    )
    assert ds.joseki.is_valid
    assert all(m in ds.joseki.molecules for m in molecules)


def test_from_cams_reanalysis_molecules_add():
    """Requested molecules are added."""
    molecules = ["H2O", "CO2", "O3"]
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        molecules=molecules,
        missing_molecules_from=joseki.make("afgl_1986-us_standard")
    )
    assert ds.joseki.is_valid
    assert all(m in ds.joseki.molecules for m in molecules)

def test_from_cams_reanalysis_molecules_invalid():
    """'missing_molecules_from' is not provided."""
    molecules = ["H2O", "CO2", "O3"]
    with pytest.raises(ValueError):
        from_cams_reanalysis(
            data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
            identifier="EAC4",
            time="2020-07-01T00:00:00",
            lat=23.0,
            lon=28.0,
            molecules=molecules,
        )

def test_from_cams_reanalysis_extrapolate_up():
    """Profile is extrapolated up."""
    zup = 120 * ureg.km
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        extrapolate={"up": {"z": zup}}
    )
    assert ds.joseki.is_valid
    assert_almost_equal(ds.z[-1], zup.m)

def test_from_cams_reanalysis_extrapolate_down():
    """Profile is extrapolated down."""
    zdown = - 0.01 * ureg.km
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        extrapolate={"down": {"z": zdown}}
    )
    assert ds.joseki.is_valid
    assert_almost_equal(ds.z[0], zdown.m)

def test_from_cams_reanalysis_extrapolate_invalid():
    """Raises when extrapolation specifications are invalid."""
    with pytest.raises(ValueError):
        from_cams_reanalysis(
            data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
            identifier="EAC4",
            time="2020-07-01T00:00:00",
            lat=23.0,
            lon=28.0,
            extrapolate={"invalid": "specifications"}
        )

def test_from_cams_reanalysis_regularize_bool():
    """Profile is regularized."""
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        regularize=True,
    )
    assert ds.joseki.is_valid
    zdiff = np.diff(ds.z.values)
    assert_allclose(zdiff, zdiff[0])

def test_from_cams_reanalysis_regularize_specs():
    """Profile is regularized."""
    num = 32
    ds = from_cams_reanalysis(
        data="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        identifier="EAC4",
        time="2020-07-01T00:00:00",
        lat=23.0,
        lon=28.0,
        regularize={"options": {"num": num}},
    )
    assert ds.joseki.is_valid
    assert ds.z.size == num

def test_get_molecule_amounts():
    """Returns mapping of H2O and O3 and their total amounts as quantities."""
    amounts = get_molecule_amounts(
        "tests/data/232f85b4-b3e9-47a7-a52d-9f955c38b9f6.nc",
        lat=23.0,
        lon=28.0,
        time="2020-07-01T00:00:00",
    )

    assert all([m in amounts for m in ["H2O", "O3"]])
    assert all(isinstance(k, pint.Quantity) for k in amounts.values())

def test_get_molecule_amounts_h2o():
    """Returns mapping of H2O and its total amount as quantity."""
    molecules = ["H2O"]
    amounts = get_molecule_amounts(
        "tests/data/232f85b4-b3e9-47a7-a52d-9f955c38b9f6.nc",
        lat=23.0,
        lon=28.0,
        time="2020-07-01T00:00:00",
        molecules=molecules,
    )

    assert list(amounts.keys()) == molecules

def test_get_molecule_amounts_invalid():
    """Raises because a requested molecule is not available in the CAMS data."""

    with pytest.raises(ValueError):
        get_molecule_amounts(
            "tests/data/232f85b4-b3e9-47a7-a52d-9f955c38b9f6.nc",
            lat=23.0,
            lon=28.0,
            time="2020-07-01T00:00:00",
            molecules=["CO2"],
        )

def test_from_archive():
    """Extracts in a temporary directory when 'extract_dir' is not provided."""
    path = from_archive(
        path="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
    )
    assert path.exists()

def test_from_archive_extract_dir():
    """Extracts in a temporary directory."""
    tmpdir = tempfile.mkdtemp()
    extract_dir = Path(tmpdir)
    path = from_archive(
        path="tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb.zip",
        extract_dir=extract_dir,
    )
    assert path.exists()

def test_to_paths_invalid():

    with pytest.raises(FileNotFoundError):
        to_paths("tests/data/97dd7a58-674a-4a1f-a92b-534cb95d07bb.zip")
        ########################^ ('a' was replaced with 'd')

def test_to_paths_directory():
    paths = to_paths("tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb/")
    assert len(paths) == 2

def test_to_paths_list():
    paths = to_paths(
        [
            "tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb/levtype_ml.nc",
            "tests/data/97da7a58-674a-4a1f-a92b-534cb95d07bb/levtype_sfc.nc"
        ]
    )
    assert len(paths) == 2

def test_to_paths_invalid_type():
    with pytest.raises(NotImplementedError):
        to_paths(42)

def test_from_cams_reanalysis_pressure_data():
    """Surface pressure data is used to rescale the pressure profile."""
    ds = from_cams_reanalysis(
        data="tests/data/25b63a29-3eab-4599-92f4-7bba42ec6add.zip",
        identifier="EAC4",
        time="2021-04-01T12:00:00",
        lat=27.5,
        lon=86.5,
        pressure_data="surface_pressure",
    )

    surface_pressure = to_quantity(ds.p.isel(z=0))
    assert np.abs(surface_pressure - 1013.25 * ureg.hPa).magnitude > 1.0



"""Test cases for the __main__ module."""
import pathlib
import typing as t

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from click.testing import CliRunner

from joseki import __main__


@pytest.fixture
def runner() -> CliRunner:
    """Fixture for invoking command-line interfaces."""
    return CliRunner()


@pytest.mark.parametrize(
    "identifier", ["afgl_1986-midlatitude_summer", "mipas_2007-midlatitude_day"]
)
def test_main_succeeds(runner: CliRunner, tmpdir: t.Any, identifier: str):
    """Exits with a status code of zero."""
    result = runner.invoke(
        __main__.main,
        [f"--identifier={identifier}", f"--file-name={tmpdir / 'ds.nc'}"],
    )
    assert result.exit_code == 0


def test_main_open_data_set(runner: CliRunner, tmpdir: t.Any):
    """Returns a xarray.Dataset."""
    path = pathlib.Path(tmpdir / "ds.nc")
    runner.invoke(
        __main__.main,
        ["--identifier=afgl_1986-tropical", f"--file-name={path}"],
    )
    assert isinstance(xr.open_dataset(path), xr.Dataset)


@pytest.fixture
def altitudes_path(tmpdir: t.Any) -> pathlib.Path:
    """Fixture for altitudes file path."""
    df = pd.DataFrame(np.linspace(0, 120, 121), columns=["z"])
    path = pathlib.Path(tmpdir, "z.csv")
    df.to_csv(path)
    return path


def test_main_altitude_path(
    runner: CliRunner, tmpdir: t.Any, altitudes_path: pathlib.Path
):
    """Exits with a status code of zero when --altitudes option is used."""
    result = runner.invoke(
        __main__.main,
        [
            "--identifier=afgl_1986-tropical",
            f"--file-name={tmpdir / 'ds.nc'}",
            f"--altitudes={altitudes_path}",
        ],
    )
    assert result.exit_code == 0


def test_main_represent_in_cells(runner: CliRunner, tmpdir: t.Any):
    """Exits with a status code of zero with option --represent-in-cells."""
    result = runner.invoke(
        __main__.main,
        [
            "--identifier=afgl_1986-tropical",
            f"--file-name={tmpdir / 'ds.nc'}",
            "--represent-in-cells",
        ],
    )
    assert result.exit_code == 0


def test_main_interp_method(runner: CliRunner, tmpdir: t.Any):
    """Exits with a status code of zero with option --p-interp-method."""
    result = runner.invoke(
        __main__.main,
        [
            "--identifier=afgl_1986-tropical",
            f"--file-name={tmpdir / 'ds.nc'}",
            "--p-interp-method=quadratic",
        ],
    )
    assert result.exit_code == 0


def test_main_conserve_column(runner: CliRunner, tmpdir: t.Any):
    """Exits with a status code of zero with option --conserve-column"""
    result = runner.invoke(
        __main__.main,
        [
            "--identifier=afgl_1986-tropical",
            f"--file-name={tmpdir / 'ds.nc'}",
            "--represent-in-cells",
            "--conserve-column",
        ],
    )
    assert result.exit_code == 0



"""Test cases for the core module."""
import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_approx_equal

from joseki import unit_registry as ureg
from joseki.core import make, open_dataset, load_dataset, merge, identifiers


def test_make():
    """Returns xr.Dataset."""
    assert isinstance(make(identifier="afgl_1986-tropical"), xr.Dataset)


def test_make_represent_in_cells():
    """Returned dataset has dimensions zbv and z and data variable z_bounds."""
    ds = make(
        identifier="afgl_1986-tropical",
        represent_in_cells=True,
    )
    for dim in ["zbv", "z"]:
        assert dim in ds.dims
    assert "z_bounds" in ds.data_vars


def test_make_altitudes():
    """Assigns dataset' altitude values from file."""
    z = np.linspace(0, 120, 121) * ureg.km
    ds = make(identifier="afgl_1986-tropical", z=z)
    assert np.allclose(ds.z.values, z.m_as(ds.z.attrs["units"]))


@pytest.mark.parametrize(
    "identifier",
    [
        "afgl_1986-tropical",
        "afgl_1986-midlatitude_summer",
        "afgl_1986-midlatitude_winter",
        "afgl_1986-subarctic_summer",
        "afgl_1986-subarctic_winter",
        "afgl_1986-us_standard",
    ],
)
def test_make_additional_molecules_false(identifier: str):
    """Additional molecules not included when additional_molecules=False."""
    ds = make(identifier=identifier, additional_molecules=False)
    assert len(ds.joseki.molecules) == 7


@pytest.mark.parametrize(
    "identifier",
    [
        "afgl_1986-tropical",
        "afgl_1986-midlatitude_summer",
        "afgl_1986-midlatitude_winter",
        "afgl_1986-subarctic_summer",
        "afgl_1986-subarctic_winter",
        "afgl_1986-us_standard",
    ],
)
def test_make_additional_molecules_true(identifier: str):
    """Additional molecules are included when additional_molecules=True."""
    ds = make(identifier=identifier, additional_molecules=True)
    assert len(ds.joseki.molecules) == 28


def test_make_ussa_1976():
    """Returns dataset."""
    ds = make(identifier="ussa_1976")
    assert ds.joseki.is_valid


def test_make_ussa_1976_z():
    """Returns dataset."""
    z = np.linspace(0.0, 100) * ureg.km
    ds = make(identifier="ussa_1976", z=z)
    assert ds.joseki.is_valid


def test_make_conserve_column_1():
    """Column densities are conserved when conserve_column is True."""
    ds0 = make(identifier="afgl_1986-us_standard")
    ds1 = make(
        identifier="afgl_1986-us_standard",
        z=np.linspace(0, 100, 21) * ureg.km,  # different than default
        conserve_column=True,
    )

    for m in ds0.joseki.molecules:
        assert_approx_equal(
            ds0.joseki.column_number_density[m].m,
            ds1.joseki.column_number_density[m].m,
            significant=6
        )

def test_make_conserve_column_2():
    """Column densities are conserved when conserve_column is True."""
    ds0 = make(identifier="afgl_1986-us_standard")
    ds1 = make(
        identifier="afgl_1986-us_standard",
        represent_in_cells=True,
        conserve_column=True,
    )

    for m in ds0.joseki.molecules:
        assert_approx_equal(
            ds0.joseki.column_number_density[m].m,
            ds1.joseki.column_number_density[m].m,
            significant=6
        )

def test_make_conserve_column_3():
    """Column densities are conserved when conserve_column is True."""
    ds0 = make(identifier="afgl_1986-us_standard")
    ds1 = make(
        identifier="afgl_1986-us_standard",
        z=np.linspace(0, 100, 21) * ureg.km,  # different than default
        represent_in_cells=True,
        conserve_column=True,
    )

    for m in ds0.joseki.molecules:
        assert_approx_equal(
            ds0.joseki.column_number_density[m].m,
            ds1.joseki.column_number_density[m].m,
            significant=6
        )

def test_select_molecules():
    """Returns xr.Dataset."""
    ds = make(identifier="afgl_1986-tropical", molecules=["H2O", "CO2"])
    assert ds.joseki.molecules == ["H2O", "CO2"]

def test_open_dataset(tmpdir):
    """Returns xr.Dataset."""
    ds = make(identifier="afgl_1986-tropical")
    path = tmpdir.join("test.nc")
    ds.to_netcdf(path)
    ds2 = open_dataset(path)
    assert ds2.joseki.is_valid

def test_load_dataset(tmpdir):
    """Returns xr.Dataset."""
    ds = make(identifier="afgl_1986-tropical")
    path = tmpdir.join("test.nc")
    ds.to_netcdf(path)
    ds2 = load_dataset(path)
    assert ds2.joseki.is_valid

def test_merge_1():
    ds1 = make(identifier="afgl_1986-tropical", molecules=["H2O", "CO2"])
    ds2 = make(identifier="afgl_1986-tropical", molecules=["O3"])
    ds = merge([ds1, ds2])
    assert ds.joseki.molecules == ["H2O", "CO2", "O3"]

def test_identifiers():
    """Returns list of identifiers."""
    assert isinstance(identifiers(), list)



"""Units module."""
from __future__ import annotations

from functools import singledispatch

import pint
import numpy as np
import xarray as xr
from numpy.typing import ArrayLike

ureg = pint.UnitRegistry()

pint.set_application_registry(ureg)

# define ppm ppb and ppt units
ureg.define("parts_per_million = 1e-6 = ppm = ppmv")
ureg.define('parts_per_billion = 1e-9 = ppb')
ureg.define('parts_per_trillion = 1e-12 = ppt')

# define dobson unit
ureg.define("dobson_unit = 2.687e20 * meter^-2 = du = dobson = dobson_units")


@singledispatch
def to_quantity(
    value: pint.Quantity | dict | int | float | list | np.ndarray | xr.DataArray,
    units: None | str = None,
) -> pint.Quantity:
    """Convert to a `pint.Quantity`.

    Args:
        value: Value which will be converted. If value is an `ArrayLike`, it is
            assumed to be dimensionless (unless `units` is set).
            If a `xarray.DataArray` is passed and 
            `units_error` is `True`, it is assumed to have a `units` key in
            its `attrs` field; otherwise, it is assumed to be dimensionless.
        units: Units to assign. If `None`, the units are inferred from the
            `value` argument.

    Returns:
        The corresponding quantity.
    
    Notes:
        This function can also be used on DataArray and Dataset coordinate 
        variables.
    """
    raise NotImplementedError

@to_quantity.register(pint.Quantity)
def _(
    value, 
    units: None | str = None,
) -> pint.Quantity:
    return value.to(units) if units else value

@to_quantity.register(dict)
def _(
    value,
    units: None | str = None,
) -> pint.Quantity:
    return ureg.Quantity(**value).to(units) if units else ureg.Quantity(**value)

@to_quantity.register(int)
def _(
    value,
    units: None | str = None,
) -> pint.Quantity:
    return to_quantity_array_like(value, units)

@to_quantity.register(float)
def _(
    value,
    units: None | str = None,
) -> pint.Quantity:
    return to_quantity_array_like(value, units)

@to_quantity.register(list)
def _(
    value,
    units: None | str = None,
) -> pint.Quantity:
    return to_quantity_array_like(np.array(value), units)

@to_quantity.register(np.ndarray)
def _(
    value,
    units: None | str = None,
) -> pint.Quantity:
    return to_quantity_array_like(value, units)

@to_quantity.register(xr.DataArray)
def _(
    value: xr.DataArray,
    units: None | str = None,
) -> pint.Quantity:
    try:
        da_units = value.attrs["units"]
    except KeyError as e:
        if units is None:
            raise ValueError(
                "this DataArray has no 'units' metadata field"
            ) from e
        else:
            return value.values * ureg(units)
    else:
        q = value.values * ureg(da_units)
        if units is not None:
            return q.to(units)
        else:
            return q
    

def to_quantity_array_like(
    value: ArrayLike,
    units: None | str = None,
) -> pint.Quantity:
    try:
        magnitude = np.array(value)
    except ValueError as e:  # pragma: no cover
        raise TypeError(
            f"Could not convert this type ({type(value)}) to a numpy array"
        ) from e
    else:
        if units is not None:
            return magnitude * ureg(units)
        else:
            return magnitude * ureg.dimensionless


"""Utilities to create a profile from 
[reanalysis datasets](https://confluence.ecmwf.int/display/CKB/CAMS%3A+Reanalysis+data+documentation)
([Inness+2019](bibliography.md#Inness+2019)) as provided by the
Copernicus Atmospheric Monitoring Service (CAMS).
"""
from __future__ import annotations

import datetime
import importlib_resources
import logging
import os
import tempfile
import zipfile
from pathlib import Path
from os import PathLike

import numpy as np
import pandas as pd
import pint
import xarray as xr

from ..units import ureg, to_quantity
from .util import mass_fraction_to_mole_fraction3, to_m_suffixed_data
from .schema import schema
from .core import DEFAULT_METHOD
from .core import regularize as _regularize, extrapolate as _extrapolate

logger = logging.getLogger(__name__)

# TODO: complete this mapping
# Mapping of CAMS specific ratio data variables names to molecule formula.
MOLECULE = {
    "co": "CO",
    "q": "H2O",
    "go3": "O3",
    "hno3": "HNO3",
    "no2": "NO2",
    "no": "NO",
    "oh": "OH",
    "o3": "O3",
    "so2": "SO2",
    "ch4": "CH4",
    "co2": "CO2",
}

# TODO: complete this mapping
# Mapping of total amounts variable names to molecule formula.
MOLECULE_TOTALS = {
    "tcco": "CO",
    "tc_ch4": "CH4",
    "tcno2": "NO2",
    "tc_no": "NO",
    "gtco3": "O3",
    "tcso2": "SO2",
    "tcwv": "H2O",
    "tcch4": "CH4",
    "tcco2": "CO2",
}


def from_cams_reanalysis(
    data: PathLike | list[PathLike],
    identifier: str,
    time: str | datetime.datetime | np.datetime64,
    lat: float | pint.Quantity,
    lon: float | pint.Quantity,
    molecules: list[str] | None = None,
    missing_molecules_from: xr.Dataset | None = None,
    extrapolate: dict | None = None,
    regularize: bool | dict | None = None,
    extract_dir : Path | None = None,
    pressure_data : str | None = "surface_pressure",
) -> xr.Dataset:
    """Convert CAMS reanalysis data to a profile.

    Args:
        data: Path to CAMS data archive file or list of paths to CAMS datasets 
            or path to directory where CAMS datasets are located.
        identifier: `'EAC4'` for a *CAMS reanalysis of atmospheric composition* 
            dataset, `'EGG4'` for a *CAMS global greenhouse gas reanalysis*
            dataset. This is used to populate the dataset `reference`, `url` 
            and `urldate` attributes.
        time: Time.
        lat: Latitude [degrees].
        lon: Longitude [degrees].
        molecules: List of molecules to include in the profile. If this list
            includes molecules that are not in the CAMS data.
            If None, all molecules are included.
        missing_molecules_from: Reference dataset to use for missing molecules.
        extrapolate: Specifications to extrapolate the CAMS data to lower/upper 
            altitudes. Accepted keys: 'up', 'down' and 'method'. 
            For each key in ['up', down'], the value is another mapping with
            a required 'z' key that specifies the altitude values to extrapolate
            to and an optional 'method' key that is a dictionary that maps 
            data variables and extrapolation methods to use.
        regularize: Specifications to regularize the altitude grid.
            'method': Data variable and interpolation method mapping. When the
                interpolation method is not specified for a data variable, the
                default method is used.
            'conserve_column': If True, ensure that column densities are 
                conserved.
            'options': Regularization options.
        extract_dir: Directory where to extract the CAMS data. If None, a
            temporary directory is used.
        pressure_data: indicates how to compute the pressure data. Either
            'model_level_60' or 'surface_pressure'. If 'model_level_60',
            the pressure data is computed from the L60 model level table.
            If 'surface_pressure', the pressure data is computed by rescaling
            the L60 model level table with the surface pressure from the CAMS
            surface dataset.
    
    Raises:
        ValueError: If the dataset is not a multi-level dataset or if the 
            archive does not contain a multi-level dataset.
        ValueError: If molecules are requested that are not in the dataset
            and `missing_molecules_from` is not specified.
        ValueError: If `pressure_data` is set to `'surface_pressure'` but the
            CAMS surface dataset is not provided.

    Returns:
        Profile dataset.

    Notes:
        Supports only the 
        [CAMS reanalysis datasets](https://confluence.ecmwf.int/display/CKB/CAMS%3A+Reanalysis+data+documentation).
        Furthermore, the dataset must be a multi-level dataset, (identified as 
        such with the presence of a 'level' coordinate).

        There are currently some assumptions/limitations with respect to the 
        conversion of CAMS 'model levels' into altitude values:
        * The model level to altitude conversion is based on the 
          [L60 model level definitions](https://confluence.ecmwf.int/display/UDOC/L60+model+level+definitions)
          where the "n" and "Geometric Altitude [m]" table columns provide
          the relationship between model levels and altitude values.
        * This conversion thus assumes that the local ground surface is at
          mean sea level, as in the U.S. Standard Atmosphere, 1976 model.
          This means that significant deviations occur when the local ground
          surface is not at mean sea level.
        * The CAMS data does not provide the air pressure on model levels.
          The pressure data injected in the profile build here is also based
          on the *L60 model level table*, i.e. the U.S. Standard Atmosphere, 
          1976 model. This means that the pressure profile is always the same.
    """
    longitude = to_quantity(lon, units="degree")
    latitude = to_quantity(lat, units="degree")

    # Open datasets (we sort by time because CAMS data is not always sorted by 
    # time and this is required to select at a timestamp)
    datasets = [
        xr.open_dataset(p).sortby("time") for p in to_paths(
            cams_data=data,
            extract_dir=extract_dir,
        )
    ]
    logger.debug("Opened %d datasets.", len(datasets))

    level_dataset, surface_dataset = identify(datasets)

    if level_dataset is None:  # pragma: no cover
        raise ValueError("Could not find a multi-level dataset.") 

    # Select in time and space
    interpolated = interp_time_space(
        level_dataset,
        time,
        longitude.m,
        latitude.m,
    )

    pressure_data = "surface_pressure" if pressure_data is None else pressure_data

    # Read surface pressure if available
    if surface_dataset is not None:
        surface_selected = interp_time_space(
            surface_dataset,
            time,
            longitude.m,
            latitude.m,
        )
        if "sp" in surface_selected:
            surface_pressure = to_quantity(surface_selected["sp"])
        else:
            if pressure_data == "surface_pressure":  # pragma: no cover
                raise ValueError(
                    "Could not found a 'sp' surface pressure data variable "
                    "in the surface dataset."
                    "This is required when the option 'pressure_data' is set "
                    "to 'surface_pressure'."
                )
            surface_pressure = None  # pragma: no cover
    else:
        if pressure_data == "surface_pressure":  # pragma: no cover
            raise ValueError(
                "The CAMS surface dataset could not be found. "
                "As a result, the surface pressure cannot be fetched."
                "This is required when the option 'pressure_data' is set to "
                "'surface_pressure'."
            )
        else:
            surface_pressure = None  # pragma: no cover

    if surface_pressure is not None:  # pragma: no cover
        logger.info("Surface pressure: %s", f"{surface_pressure:~P}")

    # Translate model levels into altitude values (and add pressure data)
    altitude_dataset = model_level_to_altitude(
        interpolated,
        pressure_data=pressure_data,
        surface_pressure=surface_pressure,
    )

    # Create a data structure storing mole fractions
    data = mole_fractions(altitude_dataset)

    # Drop extra molecules
    if molecules:
        data = data.drop_vars(
            [
                xm for xm in data
                if xm.startswith("x_") and xm not in [
                    f"x_{m}" for m in molecules
                ]
            ]
        )

    # compute missing molecules
    if molecules:
        missing_molecules = [m for m in molecules if f"x_{m}" not in data]
        logger.debug(
            "The following requested molecules are not in the CAMS data: %s",
            ", ".join(missing_molecules),
        )

        if missing_molecules and not missing_molecules_from:
            raise ValueError(
                "Missing molecules were requested but no reference dataset "
                "was provided."
            )
    
    # add missing molecules
    if missing_molecules_from is not None:
        _data = missing_molecules_from.drop_vars([
            v for v in missing_molecules_from.data_vars
            if not v.startswith("x_")
        ])

        _all_molecules = [x[2:] for x in _data.data_vars]
        _drop_molecules = [
            m for m in _all_molecules
            if m not in missing_molecules
        ]
        _missing_data = _data.drop_vars([
            f"x_{m}" for m in _drop_molecules
        ])
    
        logger.debug("data z units: %s", data.z.units)
        # missing molecule data must be on the same altitude grid
        _missing_data = _missing_data.interp(
            z=data.z.values,
        )

        # missing molecule data can be added
        data = data.assign(
            **{
                f"x_{m}": _missing_data[f"x_{m}"]
                for m in missing_molecules
            }
        )

    # add pressure and temperature data
    data = data.assign(
        {
            "p": (
                    "z",
                    altitude_dataset["p"].values,
                    {"units": altitude_dataset["p"].units},
                ),
            "t": (
                    "z",
                    altitude_dataset["t"].values,
                    {"units": altitude_dataset["t"].units},
                ),
        }
    )


    # Dataset attributes
    INSTITUTION = "European Centre for Medium-Range Weather Forecasts (ECMWF)"
    SOURCE = (
        "4DVar data assimilation in CY42R1 of ECMWF’s Integrated Forecast "
        "System (IFS)"
    )

    utcnow = datetime.datetime.utcnow().replace(microsecond=0).isoformat()

    # This comment follows the Copernicus Products Licence Agreement requirements:
    # https://cds.climate.copernicus.eu/api/v2/terms/static/licence-to-use-copernicus-products.pdf
    COMMENT = (
        "Downloaded from the Copernicus Atmosphere Monitoring Service (CAMS) "
        "Atmosphere Data Store (ADS) (<URL to dataset overview page>)"
        "Contains modified Copernicus Atmosphere Monitoring Service information "
        "2023. Neither the European Commission nor ECMWF is responsible for any "
        "use that may be made of the Copernicus information or data it contains."
    )

    # Citing
    # https://confluence.ecmwf.int/display/CKB/CAMS%3A+Reanalysis+data+documentation#CAMS:Reanalysisdatadocumentation-HowtocitetheCAMSGlobalReanalysis

    # Overview pages URLs
    URL = {
        "EAC4": "https://ads.atmosphere.copernicus.eu/cdsapp#!/dataset/cams-global-reanalysis-eac4?tab=overview",
        "EGG4": "https://ads.atmosphere.copernicus.eu/cdsapp#!/dataset/cams-global-ghg-reanalysis-egg4?tab=overview",
    }
    acknowledgment = (
        f"Downloaded from the Copernicus Atmosphere Monitoring Service (CAMS) "
        f"Atmosphere Data Store (ADS) ({URL[identifier]})"
    )

    # Published / web-based references
    REFERENCE = {
        "EAC4": "https://doi.org/10.5194/acp-19-3515-2019",
        "EGG4": "https://confluence.ecmwf.int/display/CKB/CAMS%3A+Reanalysis+data+documentation"
    }

    URL_DATE = {
        "EAC4": "2023-06-16",
        "EGG4": "2023-06-16",
    }

    ds = schema.convert(
        data_vars={
            "p": to_quantity(data.p),
            "t": to_quantity(data.t),
            **{
                xm: to_quantity(data[xm])
                for xm in data if xm.startswith("x_")
            },
        },
        coords={
            "z": to_quantity(data.z),
        },
        attrs={
            "title": "CAMS global reanalysis (EAC4) atmospheric profile",
            "institution": INSTITUTION,
            "source": SOURCE,
            "history": f"{utcnow} - CAMS data converted to Joseki format",
            "references": REFERENCE[identifier],
            "comment": COMMENT,
            "acknowledgment": acknowledgment,
            "url": URL[identifier],
            "urldate": URL_DATE[identifier],
            "latitude": f"{latitude:~}",
            "longitude": f"{longitude:~}",
            "Conventions": "CF-1.10",
        }
    )

    if extrapolate:
        #z = to_quantity(ds.z)
        extrapolate_up = extrapolate.get("up", None)
        extrapolate_down = extrapolate.get("down", None)

        if extrapolate_up is None and extrapolate_down is None:
            raise ValueError("Either 'up' or 'down' must be specified.")

        if extrapolate_up:
            z_up = extrapolate_up.get("z", None)
            method_up = extrapolate_up.get("method", {"default": "nearest"})
            ds = _extrapolate(
                ds=ds,
                direction="up",
                z_extra=z_up,
                method=method_up,
            )
        if extrapolate_down:
            z_down = extrapolate_down.get("z", None)
            method_down = extrapolate_down.get("method", {"default": "linear"})
            ds = _extrapolate(
                ds=ds,
                direction="down",
                z_extra=z_down,
                method=method_down,
            )
    else:
        ds = ds

    if regularize:
        z = to_quantity(ds.z)
        default_num = int((z.max() - z.min()) // np.diff(z).min())
        if isinstance(regularize, bool):
            regularize = {}
        ds = _regularize(
            ds=ds,
            method=regularize.get("method", DEFAULT_METHOD),
            conserve_column=regularize.get("conserve_column", False),
            options=regularize.get('options', {"num": default_num}),
        )
    else:
        ds = ds
    
    return ds


# TODO: add support for .tar.gz archives
def from_archive(path: Path, extract_dir : Path | None = None) -> Path:
    """Extract archive file and return directory with extracted files.

    Args:
        path: Path to the archive file.
        extract_dir: Directory where to place the extracted files. If unset,
            a temporary directory is created.
    Returns:
        Directory with extracted files.
    """
    
    if extract_dir is None:
        tmpdir = tempfile.mkdtemp()
        extract_dir = Path(tmpdir)

    logger.info("Extracting archive to %s", extract_dir.absolute())
    with zipfile.ZipFile(path, 'r') as zip_ref:
        zip_ref.extractall(extract_dir)
        extracted_files = [_ for _ in extract_dir.glob("*.nc")]

    logger.debug("Completed extraction.")
    for file in extracted_files:
        filesize = os.path.getsize(file) * ureg.byte
        filesize_repr = f"{filesize.to('gigabyte'):.3f~}"
        logger.debug("File: %s. Size: %s", file, filesize_repr)

    return extract_dir


def to_paths(
    cams_data: PathLike | list[PathLike],
    extract_dir : Path | None = None,
) -> list[Path]:
    """
    Convert input to paths to CAMS dataset(s).

    Args:
        cams_data: path(s) to NetCDF file(s) (.nc), path to directory with NetCDF 
            files or path to archive file (.zip).
        extract_dir: Directory where to place the extracted files. If unset,
            a temporary directory is created.
    
    Returns:
        List of paths.
    
    Notes:
        `path` could be the path to the CAMS archive file (.zip), in which 
        case the archive file is extracted in a temporary folder and the 
        ectracted files loaded into datasets.
        `path` could also be the path to the directory with the extracted 
        files (.nc).
        Finally, `path` could be the path to an individual NetCDF file.
    """
    _dir = None

    # Single path
    if isinstance(cams_data, (str, Path)):
        cams_data = Path(cams_data)
        if not cams_data.exists():
            raise FileNotFoundError(
                f"File or directory not found: {cams_data.absolute()}"
            )

        # File
        if cams_data.is_file():

            # Archive file
            if cams_data.name.endswith(".zip"):
                _dir = from_archive(path=cams_data, extract_dir=extract_dir)
                filepaths = [_ for _ in _dir.glob("*.nc")]

            # NetCDF file
            elif cams_data.name.endswith(".nc"):
                filepaths = [cams_data]
            
            else:  # pragma: no cover
                raise NotImplementedError(
                    f"Expected a path name ending with '.zip' or with '.nc' "
                    f"(got {cams_data})."
                )
        
        # Directory
        else:
            filepaths = [_ for _ in cams_data.glob("*.nc")]

    # List of paths
    elif isinstance(cams_data, list):
        filepaths = [Path(x) for x  in cams_data]
    
    else:
        raise NotImplementedError(
            f"Expected type for 'path' to be either a Path or a list of Path "
            F"(got {type(cams_data)})."
        )
    
    logger.info("filepaths=%s", [x.absolute() for x in filepaths])

    return filepaths 


def identify(datasets: list) -> tuple:
    """Identify the level and surface datasets from a list of datasets.

    Args:
        datasets: List of datasets (1 or 2).
    
    Returns:
        Tuple of level and surface datasets.
    """
    try:
        assert len(datasets) in [1, 2]
    except AssertionError as e:  # pragma: no cover
        msg = f"Expected 1 or 2 datasets, got {len(datasets)}"
        logger.critical(msg)
        raise ValueError(msg) from e

    # look for the model level dataset ('level' coordinate)
    level_dataset = None
    surface_dataset = None
    for dataset in datasets:
        if "level" in dataset.coords:
            level_dataset = dataset
        else:
            surface_dataset = dataset
    
    return level_dataset, surface_dataset


def interp_time_space(
    ds: xr.Dataset,
    datetime: str | datetime.datetime | np.datetime64,
    lon: float,
    lat: float,
):
    """Interpolate a CAMS dataset at specific datetime, longitude and latitude.
    
    Args:
        ds: CAMS Dataset.
        datetime: Datetime.
        lon: Longitude [degrees].
        lat: Latitude [degrees].

    Notes:
        The dataset is interpolated using the linear method.
    """
    return ds.interp(
        time=datetime,
        method="linear",
        kwargs={"bounds_error": True},
    ).interp(
        latitude=lat,
        longitude=lon,
        method="linear",
        kwargs={"bounds_error": True},
    )


def getpath(filename: str) -> Path:
    """
    Get path for data file.

    Args:
        filename: File name.

    Returns:
        Path to data file.
    """
    return importlib_resources.files("joseki.data.ecmwf").joinpath(filename)


def rescale_pressure_profile(
    ds: xr.Dataset,
    surface_pressure: pint.Quantity
) -> xr.Dataset:
    """Rescale pressure profile with surface pressure.
    
    Args:
        ds: Dataset with pressure profile.
        surface_pressure: Surface pressure.
    
    Returns:
        Dataset with rescaled pressure profile.

    Notes:
        The pressure profile is multiplied by a scaling factor such that
        the surface pressure matches the provided surface pressure.
    """
    logger.debug("Rescaling pressure data with surface pressure")
    ds_surface_pressure = to_quantity(ds.p.sel(z=0, method="nearest"))
    logger.debug("Surface pressure: %s", ds_surface_pressure)
    scaling_factor = (
        surface_pressure / ds_surface_pressure
    ).m_as("dimensionless")
    logger.debug("Scaling factor: %s", scaling_factor)
    with xr.set_options(keep_attrs=True):
        ds["p"] = ds["p"] * scaling_factor
    return ds


# TODO: the model level to altitude conversion is approximated
def model_level_to_altitude(
    ds: xr.Dataset,
    pressure_data: str = "model_level_60",
    surface_pressure: pint.Quantity | None = None,
) -> xr.Dataset:
    """
    Convert model level data to altitude data and add pressure data.

    Args:
        ds: Dataset with model level data.
        pressure_data: indicates how to compute the pressure data. Either
            'model_level_60' or 'surface_pressure'. If 'model_level_60',
            the pressure data is computed from the L60 model level table.
            If 'surface_pressure', the pressure data is computed by rescaling
            the L60 model level table with the surface pressure from the CAMS
            surface dataset.
        surface_pressure: Surface pressure from the CAMS surface dataset.
            Serves to compare with L60 model level pressure data and report
            on discrepancies (this discrepancies become large over 
            high-altitude regions). If None, these discrepancies are not
            reported.
    
    Returns:
        Dataset tabulated against altitude and with pressure data.

    Notes:
        The relationship between model level and (geometric) altitude is 
        documented on the page 
        [L60 model level definitions](https://confluence.ecmwf.int/display/UDOC/L60+model+level+definitions).
        A table can be downloaded which maps the model level and (geometric) 
        altitude values. 
        This table is read it into a pandas data frame.
        From the dataframe, we create a xarray data array that tabulates the 
        model level against geometric altitude.
        We use the DataArray object to interpolate our model level datasets.
        In the process, the pressure data corresponding to the model level is 
        added.
        If pressure_data is set to 'surface_pressure', we rescale the pressure
        data with the surface pressure from the CAMS surface dataset.
        Last, we sort the data by increasing altitude value.
        This transformation is not accurate.
    """
    l60_path = getpath(filename="model_levels_60.csv")
    df = pd.read_csv(l60_path)
    level = xr.DataArray(
        df["n"].values[1:],
        dims=["z"],
        coords={
            "z": (
                "z",
                np.array(df["Geometric Altitude [m]"].values[1:], dtype=float),
                {"units": "m"},
            )
        },
    )

    # first, drop any variable called 'z' as that 
    ds = ds.drop_vars("z") if "z" in ds else ds

    interpolated = ds.interp(level=level).drop_vars("level")
    ds = interpolated.assign(
        {
            "p": (
                "z",
                np.array(df["pf [hPa]"].values[1:], dtype=float),
                {"units": "hPa"},
            ),
        }
    ).sortby("z")

    # Ensure altitude are in kilometer
    z = to_quantity(ds.z)
    ds = ds.assign({
        "z": ("z", z.m_as("km"), {"units": "km", "long_name": "altitude"})
    })

    if pressure_data == "surface_pressure":  # pragma: no cover
        rescale_pressure_profile(ds, surface_pressure)

    # Report on surface pressure discrepancy
    if surface_pressure is not None:  # pragma: no cover
        ds_surface_pressure = to_quantity(ds.p.sel(z=0, method="nearest"))
        rel_diff = (ds_surface_pressure - surface_pressure) / surface_pressure
        logger.info(
            "Dataset lower level pressure, %s",
            f"{ds_surface_pressure:~P}",
        )
        logger.info(
            "Surface pressure from CAMS surface dataset, %s",
            f"{surface_pressure:~P}",
        )
        logger.info(
            "Surface pressure discrepancy: %s",
            f"{rel_diff.m_as('1'):.1%}",
        )

    return ds


def mole_fractions(ds: xr.Dataset) -> xr.DataArray:
    """Get the mole fractions data from a dataset.

    Args:
        ds: Input dataset.

    Returns:
        Mole fractions dataset.
    
    Notes:
        Assumes that the input dataset tabulates the specific ratios as a
        function of altitude (coordinate 'z').
    """
    logger.debug("Extracting mole fraction from dataset")

    # List molecules in dataset
    molecules_cams = [m for m in ds.data_vars if m in MOLECULE]
    molecules_joseki = [MOLECULE[m] for m in molecules_cams]

    # Mass mixing ratio as a function of molecule and altitude
    y = xr.DataArray(
        data=np.atleast_2d(
            np.stack(
                [
                    ds[x].values
                    for x in molecules_cams
                ]
            )
        ),
        coords={
            "m": ("m", molecules_joseki),
            "z": ds["z"],
        }
    )

    # Convert mass mixing ratio to mole mixing ratio
    x = mass_fraction_to_mole_fraction3(y=y)

    # Re-organise the `xarray.DataArray` into a `xarray.Dataset` with one 
    # data variable per molecule mole fraction
    return to_m_suffixed_data(x)


def get_molecule_amounts(
    data: PathLike | list[PathLike],
    time: str | datetime.datetime | np.datetime64,
    lon: float | pint.Quantity,
    lat: float | pint.Quantity,
    molecules: list[str] | None = None,
    extract_dir : Path | None = None
) -> dict:
    """Get molecule amount from CAMS data.

    Args:
        data: Path to CAMS data or list of paths to CAMS data.
        time: Time.
        lon: Longitude [degrees].
        lat: Latitude [degrees].
        molecules: List of molecules to extract. If None, all molecules are
            extracted.
        extract_dir: Directory where to extract the data. If None, the data
            is extracted in a temporary directory.
    
    Raises:
        ValueError: If the molecule is not available in the dataset.
    
    Returns:
        Mapping of molecule and amount.

    Notes:
        The mapping is suitable for use with `joseki.accessor.rescale_to`.
    """
    longitude = to_quantity(lon, units="degree")
    latitude = to_quantity(lat, units="degree")

    # Open datasets (we sort by time because CAMS data is not always sorted by 
    # time and this is required to select at a timestamp)
    datasets = [
        xr.open_dataset(p).sortby("time") for p in to_paths(
            cams_data=data,
            extract_dir=extract_dir,
        )
    ]
    logger.debug("Opened %d datasets.", len(datasets))

    _, surface_dataset = identify(datasets)

    # Select in time and space
    selected = interp_time_space(surface_dataset, time, longitude.m, latitude.m)

    amount = {}
    for dv in selected.data_vars:
        try:
            m = MOLECULE_TOTALS[dv]
        except KeyError:  # pragma: no cover
            pass
        else:
            amount[m] = to_quantity(selected[dv])
    
    # check that all molecules are present
    if molecules is not None:
        for m in molecules:
            if m not in amount:
                raise ValueError(f"Molecule {m} not found in dataset.")
    
    # drop all molecules that are not requested
    if molecules is not None:
        amount = {m: amount[m] for m in molecules}
    
    return amount



"""Joseki."""
from . import accessor
from .__version__ import __version__ as __version__
from .core import make, open_dataset, load_dataset, merge, identifiers
from .units import ureg as unit_registry
from .profiles.core import interp

__all__ = [
    "accessor",
    "constants",
    "identifiers",
    "interp",
    "load_dataset",
    "make",
    "merge",
    "open_dataset",
    "unit_registry",
    "__version__",
]



"""Core module."""
from __future__ import annotations

import datetime
import logging
import os
import typing as t

import pint
import xarray as xr

from .profiles.factory import factory
from .profiles.core import (
    DEFAULT_METHOD,
    represent_profile_in_cells,
    select_molecules,
)
from .__version__ import __version__


logger = logging.getLogger(__name__)


def make(
    identifier: str,
    z: pint.Quantity | None = None,
    interp_method: t.Mapping[str, str] | None = DEFAULT_METHOD,
    represent_in_cells: bool = False,
    conserve_column: bool = False,
    molecules: t.List[str] | None = None,
    **kwargs: t.Any,
) -> xr.Dataset:
    """
    Create a profile with the specified identifier.

    Args:
        identifier: Profile identifier.
        z: Altitude values.
        interp_method: Mapping of variable and interpolation method.
        represent_in_cells: If `True`, compute the altitude layer centers and 
            interpolate the profile on the layer centers, and return the 
            interpolated profile.
        conserve_column: If `True`, ensure that column densities are conserved
            during interpolation.
        molecules: List of molecules to include in the profile.
        kwargs: Additional keyword arguments passed to the profile constructor.
    
    Returns:
        Profile as xarray.Dataset.
    """
    logger.info("Creating profile %s", identifier)
    logger.debug("z: %s", z)
    logger.debug("interp_method: %s", interp_method)
    logger.debug("represent_in_cells: %s", represent_in_cells)
    logger.debug("conserve_column: %s", conserve_column)
    logger.debug("molecules: %s", molecules)
    logger.debug("kwargs: %s", kwargs)

    profile = factory.create(identifier)

    logger.debug("exporting profile to xarray.Dataset")
    ds = profile.to_dataset(
        z=z,
        interp_method=interp_method,
        conserve_column=conserve_column,
        **kwargs,
    )

    if represent_in_cells:
        ds = represent_profile_in_cells(
            ds,
            method=interp_method,
            conserve_column=conserve_column,
        )
    
    if molecules is not None:
        ds = select_molecules(ds, molecules)
    
    return ds


def open_dataset(path: os.PathLike, *args, **kwargs) -> xr.Dataset:
    """
    Thin wrapper around `xarray.open_dataset`.
    
    Args:
        path: Path to the dataset.

    Returns:
        Profile.
    """
    return xr.open_dataset(path, *args, **kwargs)


def load_dataset(path: os.PathLike, *args, **kwargs) -> xr.Dataset:
    """
    Thin wrapper around `xarray.load_dataset`.
    
    Args:
        path: Path to the dataset.

    Returns:
        Profile.
    """
    return xr.load_dataset(path, *args, **kwargs)


def merge(
    datasets: t.Iterable[xr.Dataset],
    new_title: str | None = None,
) -> xr.Dataset:
    """
    Merge multiple profiles into a single profile.
    
    Args:
        datasets: Iterable of profiles.
        new_title: New title for the merged profile. If `None`, the title of
            the first profile is used.

    Returns:
        Merged profile.
    
    Notes:
        The first profile in the iterable is used as the base profile; when
        variables with the same name are encountered in subsequent profiles,
        the variable from the first profile is used.
    """
    merged = xr.merge(
        datasets,
        compat="override",  # pick variable from first dataset
        combine_attrs="override",  # copy attrs from the first dataset to the result
    )

    # update attributes
    now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()

    institutions = set([ds.attrs["institution"] for ds in datasets])
    sources = set([ds.attrs["source"] for ds in datasets])
    references = set([ds.attrs["references"] for ds in datasets])
    urls = set([ds.attrs["url"] for ds in datasets])
    urldates = set([ds.attrs["urldate"] for ds in datasets])
    conventions = set([ds.attrs["Conventions"] for ds in datasets])

    with_profiles = "with profile " if len(datasets) == 2 else "with profiles "
    with_profiles += ", ".join([f"'{ds.title}'" for ds in datasets[1:]])
    merged.attrs["history"] += (
        f"\n{now} - merged profile '{datasets[0].title}' {with_profiles} "
        f"joseki, version {__version__}"
    )

    merged.attrs["institution"] = "\n".join(institutions)
    merged.attrs["source"] = "\n".join(sources)
    merged.attrs["references"] = "\n".join(references)
    merged.attrs["url"] = "\n".join(urls)
    merged.attrs["urldate"] = "\n".join(urldates)
    merged.attrs["Conventions"] = "\n".join(conventions)

    if new_title is not None:  # pragma: no cover
        merged.attrs["title"] = new_title
    else:  # pragma: no cover
        merged.attrs["title"] = datasets[0].title

    return merged


def identifiers() -> t.List[str]:
    """
    List all registered profile identifiers.

    Returns:
        List of all registered profile identifiers.
    """
    return factory.registered_identifiers



import numpy as np
import pint
import pytest
import xarray as xr
from numpy.testing import assert_approx_equal, assert_allclose

from joseki import unit_registry as ureg
from joseki.core import make
from joseki.profiles.core import (
    rescale_to_column,
    interp,
    extrapolate,
    regularize,
    select_molecules,
)
from joseki.core import represent_profile_in_cells
from joseki.units import to_quantity

@pytest.fixture
def test_data_set() -> xr.Dataset:
    """Test dataset fixture."""
    return make(identifier="afgl_1986-tropical")

def test_rescale_to_column_factor_0():
    ds = make(identifier="afgl_1986-tropical")
    with xr.set_options(keep_attrs=True):
        reference = ds.copy(deep=True).assign({
            "x_H2O": ds.x_H2O * 0
        })

    rescaled = rescale_to_column(
        reference=reference,
        ds=ds
    )

    assert_allclose(rescaled.x_H2O, 0)

def test_rescale_to_column_impossible():
    reference = make(identifier="afgl_1986-tropical")
    with xr.set_options(keep_attrs=True):
        ds = reference.copy(deep=True).assign({
            "x_H2O": reference.x_H2O * 0
        })

    with pytest.raises(ValueError):
        rescale_to_column(
            reference=reference,
            ds=ds
        )

def test_interp_returns_data_set(test_data_set: xr.Dataset):
    """Returns an xarray.Dataset."""
    interpolated = interp(
        ds=test_data_set,
        z_new=np.linspace(0, 120, 121) * ureg.km,
    )
    assert isinstance(interpolated, xr.Dataset)


def test_interp_out_of_bound(test_data_set: xr.Dataset):
    """Raises when out of bounds values are provided."""
    with pytest.raises(ValueError):
        interp(ds=test_data_set, z_new=np.linspace(0, 150) * ureg.km)


def test_interp_extrapolate(test_data_set: xr.Dataset):
    """Key-word argument 'fill_value' is processed."""
    interpolated = interp(
        ds=test_data_set,
        z_new=np.linspace(0, 150) * ureg.km,
        fill_value="extrapolate"
    )
    assert interpolated.joseki.is_valid

def test_interp_bounds_error(test_data_set: xr.Dataset):
    """Key-word argument 'fill_value' is processed."""
    interpolated = interp(
        ds=test_data_set,
        z_new=np.linspace(0, 100) * ureg.km,
        bounds_error=True,
    )
    assert interpolated.joseki.is_valid


def test_represent_profile_in_cells(test_data_set: xr.Dataset):
    """Returns a xarray.Dataset object."""
    interpolated = represent_profile_in_cells(ds=test_data_set)
    assert isinstance(interpolated, xr.Dataset)


def test_represent_profile_in_cells_twice(test_data_set: xr.Dataset):
    """Running function twice has no effect."""
    ds1 = represent_profile_in_cells(ds=test_data_set)
    ds2 = represent_profile_in_cells(ds=ds1)
    assert ds1 == ds2

@pytest.mark.parametrize(
    "z_down",
    [
        -0.1 * ureg.km,
        np.linspace(-0.5, -0.1, 5) * ureg.km,
    ]
)
def test_extrapolate_down(test_data_set: xr.Dataset, z_down: pint.Quantity):
    """Extrapolate down to 100 m."""
    extrapolated = extrapolate(
        ds=test_data_set,
        z_extra=z_down,
        direction="down",
    )

    # the last value of the z coordinate should be the same as z_down
    zunits = ureg.km
    assert_approx_equal(
        to_quantity(extrapolated.z).min().m_as(zunits),
        np.atleast_1d(z_down.m_as(zunits)).min(),
    )
    assert extrapolated.joseki.is_valid

@pytest.mark.parametrize(
    "z_up",
    [
        130 * ureg.km,
        np.linspace(130, 180, 6) * ureg.km,
    ]
)
def test_extrapolate_up(test_data_set: xr.Dataset, z_up: pint.Quantity):
    """Extrapolate up to 130 km."""
    z_up = 130.0 * ureg.km
    extrapolated = extrapolate(
        ds=test_data_set,
        z_extra=z_up,
        direction="up",
    )

    # the last value of the z coordinate should be the same as z_up
    zunits = ureg.km
    assert_approx_equal(
        to_quantity(extrapolated.z).max().m_as(zunits),
        np.atleast_1d(z_up.m_as(zunits)).max(),
    )
    assert extrapolated.joseki.is_valid


def test_extrapolate_invalid_altitude(test_data_set: xr.Dataset):
    """Raises when invalid altitude is provided."""
    with pytest.raises(ValueError):
        extrapolate(
            ds=test_data_set,
            z_extra=10 * ureg.km,
            direction="down",
        )
    
    with pytest.raises(ValueError):
        extrapolate(
            ds=test_data_set,
            z_extra=60 * ureg.km,
            direction="up",
        )

def test_extrapolate_invalid_direction(test_data_set: xr.Dataset):
    """Raises when invalid direction is provided."""
    with pytest.raises(ValueError):
        extrapolate(
            ds=test_data_set,
            z_extra=130 * ureg.km,
            direction="invalid",
        )

def test_regularize(test_data_set: xr.Dataset):
    """Regularize the profile."""
    regularized = regularize(ds=test_data_set)
    zgrid = to_quantity(regularized.z)

    # grid is regular
    assert np.all(np.diff(zgrid) == zgrid[1] - zgrid[0])

def test_regularize_zstep(test_data_set: xr.Dataset):
    """Regularize the profile."""
    zstep = 0.5 * ureg.km
    regularized = regularize(
        ds=test_data_set,
        options={
            "zstep": zstep,
        }
    )

    zgrid = to_quantity(regularized.z)
    assert np.all(np.diff(zgrid) == zstep)

def test_regularize_zstep_str_invalid(test_data_set: xr.Dataset):
    """Invalid zstep raises."""
    with pytest.raises(ValueError):
        regularize(
            ds=test_data_set,
            options={
                "zstep": "invalid",
            }
        )

def test_regularize_zstep_invalid_type(test_data_set: xr.Dataset):
    """Invalid zstep raises."""
    with pytest.raises(ValueError):
        regularize(
            ds=test_data_set,
            options={
                "zstep": ["cannot be a list"],
            }
        )

def test_regularize_options_invalid(test_data_set: xr.Dataset):
    """Invalid options raises."""
    with pytest.raises(ValueError):
        regularize(
            ds=test_data_set,
            options={}
        )


def test_regularize_num(test_data_set: xr.Dataset):
    """Regularize the profile."""
    num = 543
    regularized = regularize(
        ds=test_data_set,
        options={
            "num": num,
        }
    )

    assert regularized.z.size == num


def test_select_molecules(test_data_set: xr.Dataset):
    """Select molecules."""
    selected = select_molecules(
        ds=test_data_set,
        molecules=["H2O", "CO2"],
    )

    assert "H2O" in selected.joseki.molecules
    assert "CO2" in selected.joseki.molecules
    assert "O3" not in selected.joseki.molecules

def test_select_molecules_invalid(test_data_set: xr.Dataset):
    """Raise when selected molecules are not available."""
    molecules = ["SO2", "NO2"]
    ds = test_data_set.drop_vars([f"x_{m}" for m in molecules])
    with pytest.raises(ValueError):
        select_molecules(
            ds=ds,
            molecules=["H2O", "CO2", "SO2", "NO2"],
        )
